# Copyright (c) 2020, Olav B. Fosso, NTNU
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#
# Modified by Matias Vistnes, 2021, NTNU
# 
# Smaller version of the FBS, only with the essentials

import numpy as np
from DistribObjects import State, Bus, Line

class DistLoadFlow:
    """
    Common base class Radial System's (Distribution)  Load Flow
    Input:
        BusList     - List off all Bus objects
        LineList    - List of all transmission lines objects
    Returns: None

    """

    def __init__(self, BusList: list[Bus], LineList: list[Line]):
        self.BusList = BusList
        self.LineList = LineList
        self.topology = []

        assert(not(any(bus.comp != 0 for bus in BusList)))
        assert(not(any(bus.ZIP != [0.0,0.0,1.0] for bus in BusList)))
        # The full DistLoadFlow needs to be used when there are any compensation present and when ZIP!=[0,0,1]!

    def config(self, depth=0):
        """Function for making the topology - it sets up the connection between two buses by assigned the line to the to bus
        and by preparing a list of from bus connections (branching)

        """
        def swap(line: Line, bus: Bus):
            while line != 0 and line.fbus != 1 and line.tbus != bus.busnum and line.state != State.DISABLED and bus.fromline != 1 \
                and not(any(l.state != State.DISABLED and l.tbus == line.fbus for l in self.LineList)):
                line.tbus, line.fbus = line.fbus, line.tbus
                bus = self.BusList[line.fbus - 1]
                if len(bus.tolinelist) > 1:
                    for l in bus.tolinelist[1:]: 
                        l0 = l
                        while True:
                            l1 = l0
                            l0 = self.BusList[l0.fbus-1].fromline
                            if l0 == 0 or l0 == l1: break
                            elif l0.fbus == 1:
                                swap(l, bus)
                                break
                if len(bus.tolinelist) > 0: line = bus.tolinelist[0]
                else: break

        # Delete last tree structure if any
        for bus in self.BusList:
            bus.tolinelist = []
            bus.fromline = 0
            bus.toline = 0
            bus.nextbus = []
        
        for lobj in self.LineList: # Add information to each bus of a line about which line that connects
            if lobj.state != State.DISABLED:
                self.BusList[lobj.fbus - 1].tolinelist.append(lobj)
                self.BusList[lobj.fbus - 1].toline = lobj
                self.BusList[lobj.tbus - 1].fromline = lobj if self.BusList[lobj.tbus - 1].fromline == 0 else 1 # finding double fromlines
        
        for lobj in self.LineList:
            if lobj.state != State.DISABLED:
                if (lobj.fbus > 1 and self.BusList[lobj.fbus - 1].fromline == 0):  # Identify broken chains in the grid
                    bus = self.BusList[lobj.fbus - 1]
                    line = lobj
                    assert(depth < 10) # Probably a loop or other fault if higher than 9 recursions
                    swap(line, bus)
                    self.config(depth+1)
                    return

        # Add the topology information needed to define the tree structure
        for lobj in self.LineList:
            if lobj.state != State.DISABLED:
                self.BusList[lobj.fbus - 1].nextbus.append(self.BusList[lobj.tbus - 1])  
                    # Add the next bus to the list of branches of the bus
           #     print('Line: ',lobj.fbus, lobj.tbus, self.BusList[itr].toline.fbus )

    def DistLF(self, maxit = 5, lim=0.00001, print_out=False):
        """ Solves the distribution load flow up to dV limit or until a specified number of iterations, witchever happens first.
        The two first steps are to set up additions topology information and to build the main structure
        Next, it is switched between forward sweeps (Voltage updates) and backward sweeps(load update and loss calcuation)
        """

        for i in range(maxit):
            p1,q1,p2,q2 = self.accload(self.topology)
            dV = self.UpdateVolt(self.topology)
            if print_out:
                print('Iter: ',i+1, 'Pload:', f'{p1:12.9f}', 'Qload:', f'{q1:12.9f}',
                    'Ploss:', f'{p2:12.9f}', 'Qloss:', f'{q2:12.9f}')
            if dV < lim:
                break
        if print_out:
            print('\n',"****** Load flow completed ******",'\n')

    def accload(self, topologyList):
        """Calculates the accumulated downstream active and reactive load at all buses
        and calculates the active and reactive losses of lines and make an accumulated equivalent load at the buses
        """
        pl1 = 0.0
        ql1 = 0.0
        ploss1 = 0.0
        qloss1 = 0.0

        for obj in reversed(topologyList):        # Start on last node
            bus = obj[0]
            if len(obj) > 1:
                for i in obj[1:]:       # Do for all branches at a bus
                    pl2, ql2, ploss2, qloss2 = self.accload(i)
                    pl1 += pl2      # Add accumulated powers and losses in a branch to the node where the brancing accurs.
                    ql1 += ql2
                    ploss1 += ploss2
                    qloss1 += qloss2
                bus.pblossds = ploss1
                bus.qblossds = qloss1

            if bus.state != State.FAILED:
                pl1 += bus.pload   # Add local loads
                ql1 += bus.qload
            bus.ploadds = pl1
            bus.qloadds = ql1
            if bus.fromline:
                lobj = bus.fromline
                if lobj.state != State.DISABLED:
                    pto = bus.ploadds + ploss1
                    qto = bus.qloadds + qloss1
                    lobj.ploss = lobj.r * (pto ** 2 + qto ** 2) / bus.vomag ** 2
                    lobj.qloss = lobj.x * (pto ** 2 + qto ** 2) / bus.vomag ** 2
                    ploss1 += lobj.ploss
                    qloss1 += lobj.qloss
                    bus.pblossds = ploss1
                    bus.qblossds = qloss1

        return pl1, ql1, ploss1,  qloss1            # Return the accumulated loads and losses from the current branch

    def UpdateVolt(self,topologyList):
        """Update the voltage profile based on the accumulated load on each bus
        """
        max_dV = 0.0
        for obj in topologyList:
            bus = obj[0]
            if bus.fromline:
                tline = bus.fromline
                ifr = tline.fbus - 1
                itr = tline.tbus -1

                # Update voltages and sensitivities Single Phase
                """
                Calculate the node voltages and sensitivities in the single phase case
                """

                vk2 = self.BusList[ifr].vomag ** 2
                tpload = bus.ploadds + bus.pblossds  # Find the accumulated loads and losses flowing on the branch
                tqload = bus.qloadds + bus.qblossds
                # Voltage calculation
                V = np.sqrt(vk2 - 2 * (tpload * tline.r + tqload * tline.x) + \
                           (tpload ** 2 + tqload ** 2) * (tline.r ** 2 + tline.x ** 2) / self.BusList[ifr].vomag ** 2)
                max_dV = abs(self.BusList[itr].vomag - V) if abs(self.BusList[itr].vomag - V) > max_dV else max_dV
                self.BusList[itr].vomag = V  # Update the bus voltage magnitude on the down-stream bus
 
                # Voltage angle calculation
                busvoltreal = self.BusList[ifr].vomag - (tpload * tline.r + tqload * tline.x) / self.BusList[ifr].vomag
                busvoltimag = (tqload * tline.r - tpload * tline.x) / self.BusList[ifr].vomag
                self.BusList[itr].voang = self.BusList[ifr].voang + np.arctan2(busvoltimag, busvoltreal)  # Update voltage angles

            if len(obj) > 1:
                for i in obj[1:]:         # Update voltages along the branches
                    dV = self.UpdateVolt(i)
                    max_dV = dV if dV > max_dV else max_dV
        return max_dV
    
    def BackwardSearch(self,topologyList):
        """ Visit all the nodes in a backward approach and prints the Bus name
        """
        for x in reversed(topologyList):
            if len(x) > 1:
                print('Bus' + str(x[0].busnum))
                iloop = 1
                while iloop < len(x):           # Do for all branches of a bus
                    self.BackwardSearch(x[iloop])
                    iloop += 1
            else:
                print('Bus' + str(x[0].busnum))

    def ForwardSearch(self,topologyList):
        """Visit all nodes in a forward approach and prints the Bus name
        """
        for x in topologyList:
            print(f'Bus{x[0].busnum} ', end='')
            if len(x) > 1:
                for i in x[1:]:   # Do for all branches of a bus
                    self.ForwardSearch(i)
        print('')

    def dispFlow(self,fromLine=0,toLine=0, tpres=False, short=True):
        """ Display the flow on the requested distribution lines
        """

        mainlist = []
        rowno = []

        def uij(gij,bij,tetai,tetaj):
            return (gij*np.sin(tetai-tetaj)-bij*np.cos(tetai-tetaj))

        def tij(gij,bij,tetai,tetaj):
            return (gij*np.cos(tetai-tetaj)+bij*np.sin(tetai-tetaj))

        def bij(R, X):
            return (1.0 / complex(R, X)).imag

        def gij(R, X):
            return (1.0 / complex(R, X)).real

        if toLine == 0:
            toLine = len(self.LineList)
        if tpres:
            toLine = np.minimum(fromLine + 13, toLine)
        inum = fromLine
        for line in self.LineList[fromLine:toLine]:
            ifr = line.fbus - 1
            itr = line.tbus - 1
            bsh = 0.0           # No shunts included so far
            if line.state == State.OK:
                teta1 = self.BusList[ifr].voang
                teta2 = self.BusList[itr].voang
                v1 = self.BusList[ifr].vomag
                v2 = self.BusList[itr].vomag
                b = bij(line.r,line.x)
                g = gij(line.r,line.x)

                Pfrom = g * v1 * v1 - v1 * v2 * tij(g, b, teta1, teta2)
                Pto = g * v2 * v2 - v1 * v2 * tij(g, b, teta2, teta1)
                Qfrom = -(b + bsh) * v1 * v1 - v1 * v2 * uij(g, b, teta1, teta2)
                Qto = -(b + bsh) * v2 * v2 - v1 * v2 * uij(g, b, teta2, teta1)

                if tpres == False:
                    if short:
                        print('Line', '{:3.0f}'.format(ifr+1), '->', '{:3.0f}'.format(itr+1),
                            ' P:', '{:7.4f}'.format(Pfrom), ' Q:', '{:7.4f}'.format(Qfrom))
                    else:
                        print(' FromBus :', '{:4.0f}'.format(ifr+1), ' ToBus :', '{:4.0f}'.format(itr+1),
                            ' Pfrom :', '{:7.4f}'.format(Pfrom), ' Qfrom : ', '{:7.4f}'.format(Qfrom),
                            ' Pto :', '{:7.4f}'.format(Pto), ' Qto :', '{:7.4f}'.format(Qto))

                sublist = [ifr+1, itr+1, '{:7.4f}'.format(Pfrom), '{:7.4f}'.format(Qfrom),
                        '{:7.4f}'.format(Pto), '{:7.4f}'.format(Qfrom)]
                mainlist.append(sublist)
                rowno.append('Line ' + str(inum))
                inum += 1
            else:
                print('Line', '{:3.0f}'.format(ifr+1), '->', '{:3.0f}'.format(itr+1),
                            ' State:', line.state.name)

        if tpres:
            title = 'Transmission line flow'
            colind = ['FromBus :', ' ToBus :', 'Pfrom :', ' Qfrom : ', ' Pto :', ' Qto :']
            self.tableplot(mainlist, title, colind, rowno, columncol=[], rowcol=[])

    def dispVolt(self, fromBus=0, toBus = 0, tpres=False):
        """
        Desc:    Display voltages at all buses
        Input:   tpres= False (Display in tableformat if True)
                 fromBus and toBus defines the block, If tpres=True, it will display 13 lines from fromBus
        Returns: None
        """
        mainlist = []
        rowno = []
        if toBus == 0:
            toBus = len(self.BusList)
        if tpres:
            toBus = np.minimum(fromBus + 13, toBus)

        iloop = fromBus
        print(' ')
        while iloop < toBus:
            oref = self.BusList[iloop]
            if tpres == False:
                print(' Bus no :', '{:4.0f}'.format(oref.busnum),
                      ' Vmag :', '{:7.5f}'.format(oref.vomag),
                      ' Theta :', '{:7.5f}'.format(oref.voang * 180 / np.pi))
            # Prepare for graphics presentation
            sublist = ['{:4.0f}'.format(oref.busnum),
                       '{:7.5f}'.format(oref.vomag),
                       '{:7.5f}'.format(oref.voang * 180 / np.pi)]

            mainlist.append(sublist)
            rowno.append('Bus ' + str(iloop+1))
            iloop += 1
        # Present table
        if tpres:
            title = 'Bus Voltages'
            colind = [' Bus no ', ' Vmag ', ' Theta ']
            self.tableplot(mainlist, title, colind, rowno, columncol=[], rowcol=[])

    def dispLosses(self):
        'Display total losses'
        pline = 0.0
        qline = 0.0
        for x in self.LineList:
            pline += x.ploss
            qline += x.qloss
        print('\n', 'Ploss:', pline, '   Qloss:', qline)
