# Copyright (c) 2021, Matias Vistnes, NTNU
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.

from typing import Dict
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from random import random, seed
import os

import DistLoadFlow_light as DLF
from DistribObjects import Component, State, Disconnector, ComLine, ComHub, Bus, Line
from path import configure
import plotting
import stats

class System(DLF.DistLoadFlow):
	def __init__(self, ICT:bool=True):
		'''Initialize the system. 
		ICT: True, then communication ict is included in the model. '''
		# Adding time-to-failure and -repair variable to all components and calculating the time to next event
		# All components start at State.OK
		df = pd.read_excel('test_system.xlsx', sheet_name='data', usecols=range(11), engine='openpyxl')
		data = df.to_numpy()
		BusList: list[Bus] = []
		LineList: list[Line] = []
		self.busBase = []
		self.lineBase = []
		for num, p_high, q_high, p_low, q_low, fbus, tbus, r, x, betaF, state in data:
			if not(pd.isna(num)):
				if num == 1 or p_high != '-':
					self.busBase.append([(p_high, q_high),(p_low, q_low)])
					BusList.append(Bus(int(num), p_low, q_low, betaF=1/(7.295/100 /365/24), betaR=8, ICT_r=0.75))
				LineList.append(Line(int(fbus), int(tbus), r, x, betaF=betaF, betaR=6, state=State(state)))
				self.lineBase.append((int(fbus), int(tbus)))
		BusList[-1].is_slack = True
		# Set-up and first run of power flow
		self.ICT = ICT
		super().__init__(BusList, LineList)
		self.config()
		self.topology = configure(self.BusList, self.LineList, self.BusList[0])
		assert(self.systemTopologyCheck(self.topology))

		self.bus1 = [[],[]]
		# calculate summer power used for second half of year
		self.DistLF(maxit=10, print_out=False)
		self.bus1[1] = (self.BusList[0].ploadds, self.BusList[0].qloadds, self.BusList[0].pblossds, self.BusList[0].qblossds)

		# calculate winter power used for first half of year
		for i, bus in enumerate(self.BusList):
			bus.pload = self.busBase[i][0][0]
			bus.qload = self.busBase[i][0][1]
		self.DistLF(maxit=10, print_out=False)
		self.bus1[0] = (self.BusList[0].ploadds, self.BusList[0].qloadds, self.BusList[0].pblossds, self.BusList[0].qblossds)

		if self.ICT:
			self.ComHubList: list[ComHub] = []
			self.ComLineList: list[ComLine] = []
			for bus in self.BusList:
				self.ComHubList.append(ComHub(bus, betaF = 1/(0.2 /365/24), betaR = 2.5))
			for line in self.LineList:
				if line.state != State.DISABLED:
					self.ComLineList.append(ComLine(self.BusList[line.fbus-1].toComHub, self.BusList[line.tbus-1].toComHub, 
					                                betaF = 1/(0.068 /365/24), betaR = 3))
			self.DisconnectorList: list[Disconnector] = []
			betaF = 1/(0.03 /100)
			# betaF = 1/(0.03 /365/24)
			for line in self.LineList: # inherits the state of the line, important if DISABLED
				self.DisconnectorList.append(Disconnector(self.BusList[line.fbus-1], line, state=line.state, betaF=betaF, betaR=2))
				self.DisconnectorList.append(Disconnector(self.BusList[line.tbus-1], line, state=line.state, betaF=betaF, betaR=2))
      
	def simulate(self, years: int, hours: int, outfile: str, reporting: int, lineChange: bool=True):
		''' Monte Carlo simulation of the power system for years.
		Reporting to outfile + unavailability + power, set by reporting: 0=none, 1=normal, 2=full
		'''
		self.transDict: Dict[int, list[Component]] = {} # Dictionary to easily find the next component transition
		self.transDictIso: Dict[int, list[Component]] = {} # Dictionary to easily find the next disconnector transition
		Bus.failures = 0
		for bus in self.BusList:
			if bus.pload != 0 or bus.is_slack:  # Transformer on all buses with load and the slack
				self.newTrans(bus, bus.betaF, 0)
		self.disabledLines: list[Line] = [] # list of State.DISABLED lines
		Line.failures = 0
		for line in self.LineList:
			if line.state != State.DISABLED:
				self.newTrans(line, line.betaF, 0)
			else:
				self.disabledLines.append(line)
		if self.ICT:
			ComHub.failures = 0
			ComLine.failures = 0
			for c in self.ComHubList + self.ComLineList:
				if c.state != State.DISABLED:
					self.newTrans(c, c.betaF, 0)
			Disconnector.failures = 0
			for c in self.DisconnectorList:
				if c.state != State.DISABLED:
					self.newTransIso(c, c.betaF, 0)
		if not lineChange: outfile += '_noLineChange'
		if not os.path.exists('Result'): os.makedirs('Result')
		if reporting > 0: self.f = open(os.path.join('Result',f'{outfile}.txt'), "w")
		state = True # False if no nodes are isolated after a fault
		counter = [0] # counter for continuous hours with any failure of current carrying componets
		countersum = 0 # sum of counter
		self.switchedLines: list[(Line, Line)] = [] # (the faulty line, the enabled line)
		cTime = 0 # counting time for all components
		iTime = 0 # counting time for disconnectors
		powerFail = 0 # number of power interrupting components that is State.FAILED
		busFail = 0 # number of buses that is State.FAILED (except the slack-bus counted as a powerFail)
		comFail = 0 # number of communication components that is State.FAILED
		self.faultyComponents: list[Component] = [] # all faulty components at the moment

		# For index calculation
		busENS = np.zeros((years, 2, len(self.BusList), 2))
		unavailability = np.zeros((years, len(self.BusList), 2), dtype=int) # (freq,dur)
		faulty = np.zeros((len(self.BusList)), dtype=bool)
		# totalENS = 0
		# avgTotalENS = np.zeros((years))

		print('Start:')
		for year in range(years):
			yENS = busENS[year][0]
			yUnav = unavailability[year]
			self.baseCase(0)
			if len(self.switchedLines) > 0:
				self.DistLF(maxit=5, print_out=False)
			for hour in range(hours):
				if hour == int(hours/2)+1: # change to summer load at mid year
					yENS = busENS[year][1]
					self.baseCase(1)
					if len(self.switchedLines) > 0:
						self.DistLF(maxit=5, print_out=False)
				cTime += 1
				
				# Generate new states
				if cTime in self.transDict:
					state = True # such that findIsolated will be run now or next time a failure happens
					if reporting > 0:
						self.f.write(f"\ncTime {cTime:d}: year {year:6d}, hour {hour:2d}\n")
					for c in self.transDict[cTime]:
						assert(c.transition == cTime)
						if isinstance(c, Bus):
							if c.busnum == 1:
								powerFail = self.generateNextState(c, cTime, iTime, powerFail)
							else:
								busFail = self.generateNextState(c, cTime, iTime, busFail)
						elif isinstance(c, Line) or isinstance(c, Bus) or isinstance(c, Disconnector):
							powerFail = self.generateNextState(c, cTime, iTime, powerFail)
						else:
							if c.noPowerDelay:
								if reporting > 0:
									self.f.write(f'Com device repair delayed by {c.noPowerDelay} hours.\n')
								c.transition += c.noPowerDelay
								c.noPowerDelay = 0
								if c.transition in self.transDict:
									self.transDict[c.transition].append(c)
								else:
									self.transDict[c.transition] = [c]
							else:
								comFail = self.generateNextState(c, cTime, iTime, comFail)
						if reporting > 0:
							self.f.write(f"{c.tostring()}\n")
					self.transDict.pop(cTime) # deletes current time components in dict, as they have gotten new times
				
				# Power flow calculation
				if powerFail: # number of Line, Bus and Disconnector failed
					counter[-1] += 1
					countersum += 1

					# The first hour of a fault the whole system is blackouted, if the ICT system is out or disabled,
					# and always if the slackbus is FAILED
					if self.BusList[0].state == State.FAILED or (counter[-1] == 1 and (not(self.ICT) or comFail)): 
						s = (self.BusList[0].ploadds, self.BusList[0].qloadds)
						for i, bus in enumerate(self.BusList):
							yENS[i][0] += bus.pload
							yENS[i][1] += bus.qload
						for i, bus in enumerate(yUnav):
							if not faulty[i]:
								bus[0] += 1 # frequency
								faulty[i] = True # fault
							bus[1] += 1 # duration
					else:
						if state:
							iTime += 1 # disconnectors are used and they could fail
							if iTime in self.transDictIso:
								for c in self.transDictIso[iTime]:
									assert(c.transition == iTime)
									powerFail = self.generateNextState(c, cTime, iTime, powerFail)
								self.transDictIso.pop(iTime)
							
							# finds the components that are isolated from the slackbus and root ComHub
							if self.ICT:
								state = self.findIsolatedICT(self.topology, self.topology[0][0].state, State.OK) # Could be optimized with only searching from faultyComponents radially outward
							else:
								state = self.findIsolated(self.topology, self.topology[0][0].state) # Could be optimized with only searching from faultyComponents radially outward
							if not lineChange: state = False
							elif state and self.fixIsolated(reporting): # try to switch lines if at least one component is isolated
								self.DistLF(maxit=5, print_out=False)
						#self.getFlow(reporting)
						s = self.sumShedded2(yUnav, yENS, faulty)

					# The communication devices cannot be fixed without power
					for c in self.faultyComponents:
						if (isinstance(c, ComHub) and c.toBus.state != State.OK) or \
						(isinstance(c, ComLine) and c.toComHub.toBus.state != State.OK) :
							c.noPowerDelay += 1
				elif busFail: # only failure in bus affecting the LV side
					s = self.sumShedded2(yUnav, yENS, faulty)
				else:
					if counter[-1] > 0: # if the last fault was cleared this hour
						self.switchLinesBack(cTime, reporting)
						if self.ICT:
							state = self.findIsolatedICT(self.topology, self.topology[0][0].state, State.OK) # Could be optimized with only searching from faultyComponents radially outward
						else:
							state = self.findIsolated(self.topology, self.topology[0][0].state) # Could be optimized with only searching from faultyComponents radially outward
						if reporting > 0:
							self.f.write(f'Cleared all faults after continous fault(s) for {counter[-1]} hour(s).\n')
						counter.append(0)
						faulty = np.zeros((len(self.BusList)), dtype=bool) # all buses gets restarted counting
					# NB: power flow is not calculated here, assumed ok when all components are ok
					# NB: Assumes no change in load requirement per hour
					
			s = (f'\nYear {year:6d}\n_______________\n')
			if reporting > 0:
				self.f.write(s)
			print("\r", end="")
			print(year, end="")
			# totalENS += sum((x[0] for x in busENS[year][0]))+sum((x[0] for x in busENS[year][1]))
			# avgTotalENS[year] = totalENS/(year+1)
			# if year > 10 and abs(avgTotalENS[-2] - totalENS/(year+1)) < precition: 
			# 	break

		sumP = year * (self.bus1[0][0] * (int(hours/2)) + self.bus1[0][1] * (hours - int(hours/2)))
		sumQ = year * (self.bus1[1][0] * (int(hours/2)) + self.bus1[1][1] * (hours - int(hours/2)))

		s = (f'______________________________\nEnd \n'
		     f'Bus 1 load downstream winter {self.bus1[0][0]:.2f}+j{self.bus1[0][1]:.2f} for {int(hours/2)} hours.\n'
		     f'Bus 1 load downstream summer {self.bus1[1][0]:.2f}+j{self.bus1[1][1]:.2f} for {hours - int(hours/2)} hours.\n'
		     f'Total P={sumP:f}, Q={sumQ:f}\n'
		     f'In the MV system: Interruption {countersum/hours/year*100:.7f}% of the time with an '
		     f'average duration of {(countersum/(len(counter)+(counter[-1]!=0))):.2f} hours.\n'
			 f'Failures per year: Bus {Bus.failures/year:.4f}, Line {Line.failures/year:.4f}')
		if self.ICT:
			s += f', Disconnector {Disconnector.failures/year:.4f}, ComHub {ComHub.failures/year:.4f}, ComLine {ComLine.failures/year:.4f}\n'
		print(s)
		if reporting > 0:
			self.f.write(s)
			self.f.close()

		self.f = open(os.path.join('Result',f'{outfile}_unavailability.txt'), "w")
		for y, year in enumerate(unavailability):
			np.savetxt(self.f, year, fmt='%d')
			self.f.write(f'# Year {y} \n')
		self.f.close()

		self.f = open(os.path.join('Result',f'{outfile}_busENS.txt'), "w")
		for y, year in enumerate(busENS):
			for s, season in enumerate(year):
				np.savetxt(self.f, season)
				self.f.write(f'# Season {s} \n')
			self.f.write(f'# Year {y} \n')
		self.f.close()
		
		return unavailability, busENS
      
	def generateNextState(self, component: Component, cTime: int, iTime: int, failure: int):
		'''
		Calculate and set the next state of a component.
		Return True if one component is FAILED or failure is already True
		'''
		if component.state == State.FAILED or component.state == State.DISABLED:
			if component.state == State.FAILED:
				component.state = State.OK
				if isinstance(component, Disconnector):
					self.newTransIso(component, component.betaF, iTime)
				else:
					self.newTrans(component, component.betaF, cTime)
			failure -= 1
			self.faultyComponents.remove(component)
		else:
			component.state = State.FAILED
			component.__class__.failures += 1
			if isinstance(component, Bus) and component.toComHub != 0 and component.toComHub.state == State.OK:
				# use the ICT betaR if the comHub is OK
				self.newTrans(component, component.betaR*component.ICT_r, cTime)
			else:
				self.newTrans(component, component.betaR, cTime)
			failure += 1
			self.faultyComponents.append(component)
		return failure

	def findIsolated(self, topologyList: list, state: State):
		'''
		Checks all buses and lines, only electrical, 
		if they are isolated from the slack bus recursivly.
		Fix buses that no longer are isolated.
		Return: True if 
		'''
		def stateCheck(component: Component, state: State):
			'Checks the state of the component and compare it to state, to set the state accordingly.'
			if component.state == State.FAILED:
				state = State.FAILED
			elif component.state == State.OK and state == State.FAILED:
				component.state = State.ISOLATED
			elif component.state == State.ISOLATED and state == State.OK:
				component.state = State.OK
			return state

		failed = False
		for x in topologyList:
			bus = x[0]
			if bus.fromline != 0:
				line = bus.fromline
				state = stateCheck(line, state)
			stateCheck(bus, state) # Bus-failure is only gives outage on that bus, except bus 1

			# If the bus is branching, recursivly call with all buses
			if len(x) > 1:
				for i in x[1:]: # not the first that is already checked
					self.findIsolated(i, state)
					if self.findIsolated(i, state):
						failed = True
		return failed if state == State.OK else True

	def findIsolatedICT(self, topologyList: list, state: State, comState: State):
		'''
		Checks all buses and lines, electrical and communication, 
		if they are isolated from the slack bus recursivly.
		Fix buses that no longer are isolated.

		state: State of upstream power components
		comState: State of upstream communication components
		'''
		def stateCheck(component: Component, state: State):
			'Checks the state of the component and compare it to state, to set the state accordingly.'
			if component.state == State.FAILED:
				state = State.FAILED
			elif component.state == State.OK and state == State.FAILED:
				component.state = State.ISOLATED
			elif component.state == State.ISOLATED and state == State.OK:
				component.state = State.OK
			return state

		failed = False
		for x in topologyList:
			bus = x[0]
			comHub = bus.toComHub
			if bus.fromline != 0:
				line = bus.fromline
				state = stateCheck(line, state)
				comState = stateCheck(comHub.fromComLine, comState)
				for i in line.disconnectors:
					# The disconnector is affected by the 'comState' and affects the 'state'
					stateCheck(i, comState) # The var 'comState' should NOT be updated here
					if i.state == State.FAILED:
						state = State.FAILED
			stateCheck(bus, state) # Bus-fail only gives outage on that bus, except bus 1
			if state == State.FAILED:
				comState = State.FAILED
			comState = stateCheck(comHub, comState)

			# If the bus is branching, recursivly call with all buses
			if len(x) > 1:
				for i in x[1:]: # not the first that is already checked
					if self.findIsolatedICT(i, state, comState):
						failed = True
		return failed if state == State.OK else True

	def fixIsolated(self, reporting: int):
		'''Changes the topology such that none of the buses are isolated (best case).
		Only changes one line, and the first line possible is picked, then calls itself
		if more buses are isolated.

		Return: True if at least two lines are switched with each other
		'''
		for line in self.disabledLines:
			assert(line.state == State.DISABLED)
			if not(any((line == y[0] or line == y[1]) for y in self.switchedLines)) and \
				( (self.BusList[line.fbus-1].state == State.OK and self.BusList[line.tbus-1].state == State.ISOLATED) or \
				  (self.BusList[line.tbus-1].state == State.OK and self.BusList[line.fbus-1].state == State.ISOLATED) ):
					bus = line.tbus if self.BusList[line.fbus-1].state == State.OK else line.fbus
					for l in [self.BusList[bus-1].fromline] + self.BusList[bus-1].tolinelist:
						while l != 0:
							if l.state == State.FAILED or any(i.state == State.FAILED for i in l.disconnectors):
								if any((line == y[0] or line == y[1]) for y in self.switchedLines):
									break
								line.state = State.OK
								if l.state != State.FAILED:
									self.transDict[l.transition].remove(l)
								l.state = State.DISABLED
								if self.ICT:
									for i in line.disconnectors:
										i.state = State.OK
									for i in l.disconnectors:
										if i.state != State.FAILED:
											self.transDictIso[i.transition].remove(i)
										i.state = State.DISABLED
								if self.BusList[line.tbus-1].state == State.OK:
									line.tbus, line.fbus = line.fbus, line.tbus
								if self.BusList[line.tbus-1].fromline != 0 and self.BusList[line.tbus-1].state != State.OK:
									line0 = self.BusList[line.tbus-1].fromline
									while line0 != 0 and line0.state == State.ISOLATED:
										line0.tbus, line0.fbus = line0.fbus, line0.tbus
										line0 = self.BusList[line0.tbus-1].fromline

								self.switchedLines.append((l, line))
								self.disabledLines.remove(line)
								if reporting > 0:
									self.f.write(f'Switched lines {self.switchedLines[-1][0].tostring()} and {self.switchedLines[-1][1].tostring()}\n')
								# print(f'Switched lines {self.switchedLines[-1][0].tostring()} and {self.switchedLines[-1][1].tostring()}\n')
								self.config()
								self.topology = configure(self.BusList, self.LineList, self.BusList[0])
								assert(self.systemTopologyCheck(self.topology))
								if self.ICT:
									self.findIsolatedICT(self.topology, self.topology[0][0].state, State.OK)
								else:
									self.findIsolated(self.topology, self.topology[0][0].state)
								for bus in self.BusList:
									if bus.state == State.ISOLATED:
										self.fixIsolated(reporting)
										break
								return True
							l = self.BusList[l.fbus-1].fromline
		return False

	def switchLinesBack(self, cTime: int, reporting: int):
		'Switch the lines back to base case.'
		for lines in self.switchedLines:
			if lines[0].state != State.OK:
				lines[0].state = State.OK
				self.newTrans(lines[0], lines[0].betaF, cTime)
			lines[1].state = State.DISABLED
			if self.ICT:
				for i in lines[0].disconnectors:
					if i.state == State.DISABLED:
						i.state = State.OK
						self.newTransIso(i, i.betaF, cTime)
				for i in lines[1].disconnectors:
					i.state = State.DISABLED
			self.disabledLines.append(lines[1])
			if reporting > 0:
				self.f.write(f'Un-Switched lines {self.switchedLines[-1][0].tostring()} and {self.switchedLines[-1][1].tostring()}\n')
		for i, line in enumerate(self.LineList):
			line.fbus = self.lineBase[i][0]
			line.tbus = self.lineBase[i][1]
		self.switchedLines = []
		self.config()
		self.topology = configure(self.BusList, self.LineList, self.BusList[0])
		assert(self.systemTopologyCheck(self.topology))

	def newTrans(self, component: Component, t: float, cTime: int):
		'Calculates the new transition time for the component and adds it to the correct dict.'
		component.transition = int(-t * np.log(random())) + cTime + 1 # random(): Random number between 0 and 1
		if component.transition in self.transDict:
			self.transDict[component.transition].append(component)
		else:
			self.transDict[component.transition] = [component]

	def newTransIso(self, i: Disconnector, t: float, iTime: int):
		'Calculates the new transition time for the disconnector and adds it to the correct dict.'
		i.transition = int(-t * np.log(random())) + iTime + 1 # random(): Random number between 0 and 1
		if i.transition in self.transDictIso:
			self.transDictIso[i.transition].append(i)
		else:
			self.transDictIso[i.transition] = [i]
	
	def baseCase(self, season: int=0):
		'Set the load according to season: 0 = winter, 1 = summer'
		bus = self.BusList[0]
		bus.ploadds, bus.qloadds, bus.pblossds, bus.qblossds = self.bus1[season] # only values used other than pload and qload from DistLF
		for i, bus in enumerate(self.BusList):
			bus.pload = self.busBase[i][season][0]
			bus.qload = self.busBase[i][season][1]

	def sumShedded(self, reporting: int=0):
		'''Write to file the power and state at each bus.
		Summarize and return the total and the shedded load on all buses.
		'''
		sumP = 0
		sumQ = 0
		sumPshedded = 0
		sumQshedded = 0
		for bus in self.BusList:
			if bus.state == State.OK:
				sumP += bus.pload
				sumQ += bus.qload
				if reporting > 0:
					self.f.write(f'Pload:{bus.pload:7.4f} Qload:{bus.qload:7.4f} {bus.tostring()}\n')
			elif bus.state != State.DISABLED:
				sumPshedded += bus.pload
				sumQshedded += bus.qload
				if reporting > 0:
					self.f.write(f'Pshed:{bus.pload:7.4f} Qshed:{bus.qload:7.4f} {bus.tostring()}\n')
		sumP += sumPshedded
		sumQ += sumQshedded
		if reporting > 0:
			self.f.write(f'Pshed_tot:{sumPshedded:7.4f} Qshed_tot:{sumQshedded:7.4f}\n')
		return sumP, sumQ, sumPshedded, sumQshedded

	def sumShedded2(self, unavailability, power, faulty):
		'''Summarize and return the shedded load on all buses 
		that are either failed or isolated.
		'''
		sumP, sumQ = 0, 0
		for i, bus in enumerate(self.BusList):
			if bus.state == State.FAILED or bus.state == State.ISOLATED:
				if not faulty[i]:
					unavailability[bus.busnum-1][0] += 1
					faulty[i] = True
				unavailability[bus.busnum-1][1] += 1
				power[bus.busnum-1][0] += bus.pload
				power[bus.busnum-1][1] += bus.qload
				sumP += bus.pload
				sumQ += bus.qload		
		return sumP, sumQ

	def sumBranches(self, topologyList: list):
		'''Summarize load downstream on all buses.
		'''
		sumP, sumQ = 0, 0
		for x in reversed(topologyList):
			# If the bus is branching, recursivly call with all buses
			if len(x) > 1:
				for i in x[1:]: # not the first that is already checked
					P, Q = self.sumBranches(i)
					sumP += P
					sumQ += Q
			bus = x[0]
			sumP += bus.pload
			sumQ += bus.qload
			bus.ploadds = sumP
			bus.qloadds = sumQ
			# print(f'Bus{bus.busnum} pds{bus.ploadds:.4f} qloadds{bus.qloadds:.4f}')

		return sumP, sumQ

	def systemSearch(self):
		# for comHub in self.ComHubList:
		# 	print(comHub.tostring())
		for line in self.LineList:
			if line.state != State.OK: print(line.tostring())
		for bus in self.BusList:
			if bus.state != State.OK: print(bus.tostring())
	
	def systemTopologyCheck(self, topologyList: list, check: list=[]):
		"""Visit all nodes in a forward approach and 
		return True if all buses are in the topologyList.
		"""
		if check == []:
			check = [False for _ in self.BusList]
		for x in topologyList:
			if check[x[0].busnum-1]:
				return False
			check[x[0].busnum-1] = True
			if len(x) > 1:
				for i in x[1:]:   # Do for all branches of a bus
					self.systemTopologyCheck(i, check)
		if False in check:
			return False
		else:
			return True


def change(list, mode, m=2):
	if mode == 0: # incF
		for c in list: c.betaF /= m
	elif mode == 1: # decF
		for c in list: c.betaF *= m
	elif mode == 2: # incR
		for c in list:
			c.betaR *= m
	elif mode == 3: # decR
		for c in list:
			c.betaR /= m
	else: raise RuntimeError('Not an expected mode')

def cBus(system, mode, m=2):
	change(system.BusList, mode, m)
def cLine(system, mode, m=2):
	change(system.LineList, mode, m)
def cDisconnector(system, mode, m=2):
	change(system.DisconnectorList, mode, m)
def cComHub(system, mode, m=2):
	change(system.ComHubList, mode, m)
def cComLine(system, mode, m=2):
	change(system.ComLineList, mode, m)

def factorialExperimentRun(years, hours, s, lineChange=True):
	switch = { 0: cBus, 1: cLine, 2: cDisconnector, 3: cComHub, 4: cComLine }
	name = [x.__name__ for x in switch.values()]
	for ict in ['noICT', 'ICT']:
		for i in list(switch.keys()):
			for j, mode in enumerate(['incF', 'decF', 'incR', 'decR']):
				if ict == 'noICT' and i > 1: continue
				seed(s)
				system = System(ICT=(ict == 'ICT'))
				fname = f'{ict}_{years}y_{name[i]}_{mode}'
				print(fname)
				switch.get(i) (system, j)
				system.simulate(years, hours, fname, reporting=0, lineChange=lineChange)
				del system
	return name

def run(years, hours, s, ICT, lineChange=True):
	seed(s)
	system = System(ICT=ICT)
	ict = 'ICT' if ICT else 'noICT'
	fname = f's{s}_{ict}_{years}y'
	u, p = system.simulate(years, hours, fname, reporting=1, lineChange=lineChange)
	return system, u, p

# def foo():
if __name__ == "__main__":
	plotting.loadGen()
	hours = 24*365
	for years in [100, 1000, 10000]:
		for s in [67505, 10336, 33620]:
			s1, u1, p1 = run(years, hours, s, True, lineChange=True)
			s2, u2, p2 = run(years, hours, s, False, lineChange=True)
			plotting.plotAll(f's{s}_{years}y', years, p1, p2, u1, u2)
	
	plotting.plotSeeds()
	years = 10000
	s = 67505
	s1, u1, p1 = run(years, hours, s, True, lineChange=False)
	s2, u2, p2 = run(years, hours, s, False, lineChange=False)
	plotting.plotAll(f's{s}_{years}y_noLineChange', years, p1, p2, u1, u2)
	plotting.plotNoLineChange(years, f's{s}_')
	factorialExperimentRun(years, hours, s, lineChange=False)
	data = plotting.calcFactorial(years, lineChange=False)
	data = stats.import_numbers(years)
	plotting.plotFactorial(data, years)
	
	####################


# # For finding slow parts of the code
# from time import time
# import cProfile
# import pstats
# from pstats import SortKey
# # import timeit
# if __name__ == "__main__":
# 	cProfile.run('foo()', 'restats', sort='tottime')
# 	p = pstats.Stats('restats')
# 	p.strip_dirs().sort_stats('tottime').print_stats(20)
# 	p.strip_dirs().sort_stats('cumtime').print_stats(20)
