# Copyright (c) 2021, Stine Fleischer Myhre, NTNU
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.

def unique(sequence):
    """
    Return list of unique elements while preserving the order
    """
    seen = set()
    return [x for x in sequence if not (x in seen or seen.add(x))]

def configure(BusList, LineList, slack_bus):
    """Function that sets up the nested topology array and configures the radial tree according to the slack bus"""

    def get_paths(parent_bus):
        """Function that finds all downstream paths in a radial tree

        Input:
        parent_bus(Bus): Parent bus of radial tree

        Output:
        paths(list): List of all downstream paths from parent_bus

        """
        if len(parent_bus.nextbus) == 0:
            return [[[parent_bus]]]
        paths = []
        for nbus in parent_bus.nextbus:
            for path in get_paths(nbus):
                paths.append([[parent_bus]] + path)
        return paths

    def get_topology(paths):
        """Function that constructs a nested topology array

        Input:
        paths(list): List of all downstream paths from parent_bus in radial tree

        Output:
        topology(list): Nested topology list
        """
        used_buses = list()
        main_path = paths[0]
        used_buses += main_path
        topology = [main_path]
        for path in paths[1:]:
            sub_path = list()
            for bus in path:
                if bus not in used_buses:
                    sub_path.append(bus)
            used_buses += sub_path
            if sub_path != list():
                topology.append(sub_path)

        while len(topology) > 1:
            last_path = topology[-1]
            top_bus = last_path[0][0]
            for n, path in enumerate(topology[:-1]):
                for k, bus in enumerate(path):
                    if top_bus in bus[0].nextbus:
                        topology[n][k].append(last_path)
                        topology.remove(last_path)
                        break

        topology = topology[0]

        return topology

    paths = get_paths(slack_bus)
    topology = get_topology(paths)

    return topology
