# Copyright (c) 2021, Matias Vistnes, NTNU
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.

import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import os
import random
import numpy as np
import seaborn as sns
import statistics as stat
import stats
plt.style.use('seaborn-deep')

def sortAndPrintNodes(A):
	A_map = {}
	for i, a in enumerate(A):
		if a in A_map:
			A_map[a].append(i+1)
		else:
			A_map[a] = [i+1]
	for a, nodeList in sorted(A_map.items()):
		print(f'{a:.2f}: ', end='')
		for node in nodeList:
			print(node, end='')
		print('')

def calcPower(power:list):
	numBuses = len(power[0][0])
	years = len(power)
	totalENS = np.zeros((6,years)) # totP, totQ, totWinterP, totWinterQ, totSummerP, totSummerQ
	busENS = np.zeros((6,numBuses,years)) # busP, busQ, busWinterP, busWinterQ, busSummerP, busSummerQ
	AENS = np.zeros(years)
	ACCI = np.zeros(years)
	for y, year in enumerate(power):
		u = [False for _ in range(numBuses)]
		for s, season in enumerate(year):
			for b, bus in enumerate(season):
				if bus[0] > 0: u[b] = True
				for i in range(2):
					totalENS[i][y] += bus[i]
					busENS[i][b][y] += bus[i]
					if s == 0:
						totalENS[i+2][y] += bus[i]
						busENS[i+2][b][y] += bus[i]
					else:
						totalENS[i+4][y] += bus[i]
						busENS[i+4][b][y] += bus[i]
		affected = sum(u)
		AENS[y] = totalENS[0][y] / numBuses
		if affected > 0:
			ACCI[y] = totalENS[0][y] / affected
	return [totalENS, busENS, AENS, ACCI]

def runningAVGPower(totalENS, busENS, AENS, ACCI):
	# Make the indices a running average of previous years
	
	for y in range(1, len(AENS)):
		AENS[y] = (y*AENS[y-1] + AENS[y]) / (y+1)
	for y in range(1, len(ACCI)):
		if ACCI[y] != 0:
			ACCI[y] = (y*ACCI[y-1] + ACCI[y]) / (y+1)
		else:
			ACCI[y] = ACCI[y-1]

	for y in range(1, len(totalENS[0])):
		for i in range(len(totalENS)):
			totalENS[i][y] = (y*totalENS[i][y-1] + totalENS[i][y]) / (y+1)
	
	for y in range(1, len(busENS[0][0])):
		for b in range(len(busENS[0])):
			for i in range(len(busENS)):
				busENS[i][b][y] = (y*busENS[i][b][y-1] + busENS[i][b][y]) / (y+1)

	return [totalENS, busENS, AENS, ACCI]

def calcUnavailability(unavailability:list):
	numBuses = len(unavailability[0])
	years = len(unavailability)
	failureRate = np.zeros((numBuses+1, years))
	Un = np.zeros((numBuses+1, years))
	r = np.zeros((numBuses+1, years))
	saifi = np.zeros((years))
	saidi = np.zeros((years))
	caifi = np.zeros((years))
	caidi = np.zeros((years))
	asai = np.ones((years))
	for y, year in enumerate(unavailability):
		sumF = 0                              # sum of frequency of interruption
		sumD = 0                              # sum of duration of interruption
		u = [False for _ in range(numBuses)]
		for b, bus in enumerate(year):
			f = bus[0]
			d = bus[1]
			if f != 0:
				u[b] = True
			sumF += f
			sumD += d
			failureRate[b][y] = f
			Un[b][y] = d
			if f > 0:
				r[b][y] = d / f
		
		# For the whole system
		failureRate[numBuses][y] = sumF
		Un[numBuses][y] = sumD
		if sumF > 0:
			r[numBuses][y] = sumD / sumF

		saifi[y] = sumF / numBuses
		saidi[y] = sumD / numBuses
		affected = sum(u)
		if affected > 0:
			caifi[y] = sumF / affected
		if sumF > 0:
			caidi[y] = sumD / sumF
		asai[y] = 1 - sumD / (numBuses*8760)
		
	return [failureRate, Un, r, saifi, saidi, caifi, caidi, asai]

def runningAVGUnavailability(failureRate, Un, r, saifi, saidi, caifi, caidi, asai):
	def sumAvg(data: list[float], zero):
		mu = [0 for _ in range(len(data))]
		mu[0] = data[0]
		for y in range(1, len(data)):
			if not zero or data[y] != 0:
				mu[y] = (y*mu[y-1] + data[y]) / (y+1)
			else:
				mu[y] = mu[y-1]
		s = [0 for _ in range(len(data))]
		for y in range(1, len(data)):
			s[y] = ((y-1)*s[y-1] + y*mu[y-1]**2 - (y+1)*mu[y]**2 + data[y]**2) / y
		return mu, s

	def sumAvgBus(data: list[list[float]], zero):
		mu = [0 for _ in range(len(data))]
		s = [0 for _ in range(len(data))]
		for b in range(len(data)):
			mu[b], s[b] = sumAvg(data[b], zero)
		return mu, s
	
	# Make the indices a running average of previous years
	failureRate = sumAvgBus(failureRate, False)
	Un = sumAvgBus(Un, False)
	r = sumAvgBus(r, True)
	saifi = sumAvg(saifi, False)
	saidi = sumAvg(saidi, False)
	caifi = sumAvg(caifi, True)
	caidi = sumAvg(caidi, True)
	asai = sumAvg(asai, False)

	return [failureRate, Un, r, saifi, saidi, caifi, caidi, asai]

def plotPower(data1, data2, fname: str):
	xlabels = ['Total', 'Winter', 'Summer']
	figsize = (6,3)
	for i, name in enumerate(['Total ENS P','Bus ENS P','AENS','ACCI']):
		if i < 2:
			for s, season in enumerate(xlabels):
				fig = plt.figure(i, figsize=figsize)
				ax = fig.add_subplot(111, label='')
				if i < 1:
					plt.plot(range(len(data1[i][2*s])), data1[i][2*s], label="ICT")
					plt.plot(range(len(data2[i][2*s])), data2[i][2*s], label="no-ICT")
				else:
					color1 = next(ax._get_lines.prop_cycler)["color"]
					plt.bar([0], [0])
					color2 = next(ax._get_lines.prop_cycler)["color"]
					plt.bar(range(1,len(data2[i][2*s])+1), [x[-1] for x in data2[i][2*s]], color=color2, label="no-ICT")
					plt.bar(range(1,len(data1[i][2*s])+1), [x[-1] for x in data1[i][2*s]], color=color1, label="ICT")
					# plt.plot(range(1,len(data1[i][2*s])+1), [x[-1] for x in data1[i][2*s]], label="ICT", marker='o', ls='')
					# plt.plot(range(1,len(data2[i][2*s])+1), [x[-1] for x in data2[i][2*s]], label="no-ICT", marker='o', ls='')
				plt.legend()
				# plt.xlabel('Time [hours]')
				# plt.ylabel('Bus load [pu]')
				plt.tight_layout()
				plt.savefig(f'{fname}_{name}_{season}.pdf')
				plt.clf()
		else:
			plt.figure(figsize=figsize)
			plt.axhline(y=1, color='grey', linestyle='-', alpha=0.5)
			plt.plot(range(len(data1[i])), [x/data1[i][-1] for x in data1[i]], label="ICT")
			plt.plot(range(len(data2[i])), [x/data2[i][-1] for x in data2[i]], label="no-ICT")
			# print(f'ICT {name} {data1[i][-1]}')
			# print(f'No ICT {name} {data2[i][-1]}')
			plt.legend()
			plt.xlabel('Time [years]')
			plt.ylabel(name)
			plt.tight_layout()
			plt.savefig(f'{fname}_{name}.pdf')
			plt.clf()
	# plt.close('all')

def plotUnavailability(data1, data2, fname: str):
	figsize = (6,3)
	for n in range(2): # mu, s
		for i, name in enumerate(['failureRate', 'Unavailability', 'r', 'SAIFI','SAIDI','CAIFI','CAIDI','ASAI']):
			if i > 2:
				plt.figure(figsize=figsize)
				plt.axhline(y=1, color='grey', linestyle='-', alpha=0.5)
				plt.plot(range(len(data1[i][n])), [x/data1[i][n][-1] for x in data1[i][n]], label="ICT")
				plt.plot(range(len(data2[i][n])), [x/data2[i][n][-1] for x in data2[i][n]], label="no-ICT")
				plt.xlabel('Time [years]')
			elif i == 2:
				fig = plt.figure(figsize=(figsize[0]+1, figsize[1]))
				ax = fig.add_subplot(111, label='')
				color1 = next(ax._get_lines.prop_cycler)["color"]
				B = [x[-1] for x in data1[i][n][:-1]]
				plt.bar(range(1,len(data1[i][n])), B, color=color1, label="ICT")
				plt.axhline(y=stat.mean(B), color=color1, linestyle='-', alpha=0.5)
				color2 = next(ax._get_lines.prop_cycler)["color"]
				A = [x[-1] for x in data2[i][n][:-1]]
				plt.bar(range(1,len(data2[i][n])), A, color=color2, label="no-ICT")
				plt.axhline(y=stat.mean(A), color=color2, linestyle='-', alpha=0.5)
				plt.xlabel('Node')
			else:
				fig = plt.figure(figsize=(figsize[0]+1, figsize[1]))
				ax = fig.add_subplot(111, label='')
				color1 = next(ax._get_lines.prop_cycler)["color"]
				plt.bar([0], [0])
				color2 = next(ax._get_lines.prop_cycler)["color"]
				A = [x[-1] for x in data2[i][n][:-1]]
				plt.bar(range(1,len(data2[i][n])), A, color=color2, label="no-ICT")
				plt.axhline(y=stat.mean(A), color=color2, linestyle='-', alpha=0.5)
				B = [x[-1] for x in data1[i][n][:-1]]
				plt.bar(range(1,len(data1[i][n])), B, color=color1, label="ICT")
				plt.axhline(y=stat.mean(B), color=color1, linestyle='-', alpha=0.5)
				plt.xlabel('Node')
			plt.ylabel(name)
			# else: continue
			plt.legend()
			plt.tight_layout()
			if n == 0:
				plt.savefig(f'{fname}_{name}.pdf')
			else:
				plt.savefig(f'{fname}_{name}_s2.pdf')
			plt.clf()
	# plt.close('all')

def plotVS(data1, data2, name1: str, name2: str, fname: str):
	plt.figure(figsize=(3,3))
	plt.plot(data1, data2, marker='o', ls='')
	plt.xlabel(name1)
	plt.ylabel(name2)
	plt.tight_layout()
	plt.savefig(f'{fname}_VS.pdf')
	plt.close('all')

def plotHist(data1, data2, names, fname: str):
	figsize = (4,2.5)
	for i, name in enumerate(names):
		fig = plt.figure(i, figsize=figsize)
		ax = fig.add_subplot(111, label='')
		color = next(ax._get_lines.prop_cycler)["color"]
		plt.axvline(data1[i].mean(), color=color, linestyle='dashed', linewidth=1)
		ax.hist(data1[i], color=color, alpha=0.5)
		ax.hist(data1[i], label="ICT", color=color, edgecolor=color, fc="None", lw=1)
		color = next(ax._get_lines.prop_cycler)["color"]
		plt.axvline(data2[i].mean(), color=color, linestyle='dashed', linewidth=1)
		ax.hist(data2[i], color=color, alpha=0.5)
		ax.hist(data2[i], label="no-ICT", color=color, edgecolor=color, fc="None", lw=1)
		plt.legend()
		plt.tight_layout()
		# fig.patch.set_facecolor('white')
		plt.savefig(f'{fname}_{name}_hist.pdf')
		plt.clf()
	# plt.close('all')

def plotKDE(data1, data2, names, fname: str):
	figsize = (4,2.5)
	for i, name in enumerate(names):
		fig = plt.figure(i, figsize=figsize)
		ax1 = fig.add_subplot(111, label='')
		color = next(ax1._get_lines.prop_cycler)["color"]
		sns.distplot(data1[i], hist = False, kde = True, kde_kws = {'shade': True, 'linewidth': 1}, color=color)
		ax1.set_ylabel('density, ICT', color=color)
		ax2 = plt.twinx()
		color = next(ax1._get_lines.prop_cycler)["color"]
		sns.distplot(data2[i], hist = False, kde = True, kde_kws = {'shade': True, 'linewidth': 1}, color=color, ax=ax2)
		ax2.set_ylabel('density, no ICT', color=color)
		# plt.legend()
		plt.tight_layout()
		# fig.patch.set_facecolor('white')
		plt.savefig(f'{fname}_{name}_conthist.pdf')
		plt.clf()
	# plt.close('all')

def plotBox(data, title, mode = 0):
	if len(data) == 3: 
		xlabel = ['Total', 'Winter', 'Summer']
		figsize = (4,4)
	else:
		xlabel = []
		for i in ['ICT', 'no-ICT']:
			for s in ['Total', 'Winter', 'Summer']:
				xlabel.append(i + ' ' + s)
		figsize = (8,4)
	fig1 = plt.figure(1, figsize=figsize)
	if mode == 0:
		plt.boxplot([[x[-1] for x in i] for i in data], showfliers=True)
	elif mode == 1:
		plt.boxplot(data, showfliers=True)
	plt.setp(fig1.axes, xticks=[y + 1 for y in range(len(xlabel))], xticklabels=xlabel)
	plt.title(title)
	plt.tight_layout()
	plt.show()
	plt.close('all')

def calcFactorial(years, lineChange=True):
	switch = { 0: 'base', 1: 'cBus', 2: 'cLine', 3: 'cDisconnector', 4: 'cComHub', 5: 'cComLine' }
	mode = ['incF', 'decF', 'incR', 'decR']
	data = [[[] for _ in range(len(mode))] for _ in range(2)]
	for i, name in switch.items():
		for j, m in enumerate(mode):
			if i == 0: 
				fname = f'{years}y'
				s = 's67505_'
			else: 
				fname = f'{years}y_{name}_{m}'
				s = ''
			if not lineChange: fname += '_noLineChange'

			Udata = calcUnavailability(np.loadtxt(os.path.join('Result',f'{s}ICT_{fname}_unavailability.txt'), dtype=int).reshape((years, 66, 2)))
			Udata = runningAVGUnavailability(Udata[0], Udata[1], Udata[2], Udata[3], Udata[4], Udata[5], Udata[6], Udata[7])
			Pdata = calcPower(np.loadtxt(os.path.join('Result',f'{s}ICT_{fname}_busENS.txt')).reshape((years, 2, 66, 2)))
			Pdata = runningAVGPower(Pdata[0], Pdata[1], Pdata[2], Pdata[3])
			data[0][j].append([x[0][-1] for x in Udata[3:]] + [x[-1] for x in Pdata[2:]])

			if (i < 3):
				Udata = calcUnavailability(np.loadtxt(os.path.join('Result',f'{s}noICT_{fname}_unavailability.txt'), dtype=int).reshape((years, 66, 2)))
				Udata = runningAVGUnavailability(Udata[0], Udata[1], Udata[2], Udata[3], Udata[4], Udata[5], Udata[6], Udata[7])
				Pdata = calcPower(np.loadtxt(os.path.join('Result',f'{s}noICT_{fname}_busENS.txt')).reshape((years, 2, 66, 2)))
				Pdata = runningAVGPower(Pdata[0], Pdata[1], Pdata[2], Pdata[3])
				data[1][j].append([x[0][-1] for x in Udata[3:]] + [x[-1] for x in Pdata[2:]])
	
	fname = f'factorial_{years}y'
	if not lineChange: fname += '_noLineChange'
	f = open(os.path.join('Result',f'{fname}_indices.txt'), "w")
	out = [[[[] for _ in range(len(mode))] for _ in range(2)] for _ in range(7)]
	for i, name in enumerate(['SAIFI','SAIDI','CAIFI','CAIDI','ASAI','AENS','ACCI']):
		for j, m in enumerate(mode):
			A = [x[i] for x in data[0][j]]
			f.write(f'# ICT {name} {m}\n')
			np.savetxt(f, A, delimiter=',')
			if i == 4:
				A = [(1-A[i])/(1-A[0])*100-100 for i in range(1, len(A))]
			else:
				A = [A[i]/A[0]*100-100 for i in range(1, len(A))]
			out[i][0][j] = A

			B = [x[i] for x in data[1][j]]
			f.write(f'# no-ICT {name} {m}\n')
			np.savetxt(f, B, delimiter=',')
			if i == 4:
				B = [(1-B[i])/(1-B[0])*100-100 for i in range(1, len(B))]
			else:
				B = [B[i]/B[0]*100-100 for i in range(1, len(B))]
			out[i][1][j] = B
	return out

def plotFactorial(data, years, lineChange=True):
	def autolabel(rects, ax):
		"""Attach a text label above each bar in *rects*, displaying its height."""
		ylim = axs.get_ylim()
		r = abs(ylim[0]- ylim[1])
		for rect in rects:
			height = rect.get_height()
			if (height >= 0 and ylim[1] > r/5) or (height < 0 and ylim[0] > -r/5):
				ax.annotate(f'{height:.1f}', size ='small',
							xy=(rect.get_x() + rect.get_width() / 2, height),
							xytext=(0, 2),  # 2 points vertical offset
							textcoords="offset points",
							ha='center', va='bottom')
			else:
				ax.annotate(f'{height:.1f}', size ='small',
							xy=(rect.get_x() + rect.get_width() / 2, height),
							xytext=(0, -12),  # -12 points vertical offset
							textcoords="offset points",
							ha='center', va='bottom')

	mode = ['incF', 'decF', 'incR', 'decR']
	fname = f'factorial_{years}y'
	if not lineChange: fname += '_noLineChange'
	folder = os.path.join('Figures_new2',fname)
	if not os.path.exists(folder):
		os.makedirs(folder)
	fname = os.path.join(folder,fname)	
	xlabel = ['B', 'L', 'D', 'CH', 'CL']
	for i, name in enumerate(['SAIFI','SAIDI','CAIFI','CAIDI','ASAI','AENS','ACCI']):
		fig, ax = plt.subplots(2,2,figsize=(8,4))
		for j, m in enumerate(mode):
			axs = ax[j%2,int(j>1)]
			A = data[i][0][j]
			if i == 4:
				A = [(1-A[i])/(1-A[0])*100-100 for i in range(1, len(A))]
			else:
				A = [A[i]/A[0]*100-100 for i in range(1, len(A))]
			r1 = axs.bar(xlabel, A, align='edge', width=-0.8/2, label='ICT')

			B = data[i][1][j]
			if i == 4:
				B = [(1-B[i])/(1-B[0])*100-100 for i in range(1, len(B))]
			else:
				B = [B[i]/B[0]*100-100 for i in range(1, len(B))]
			r2 = axs.bar(xlabel[:2], B, align='edge', width=0.8/2, label='no-ICT')
			autolabel(r1, axs)
			autolabel(r2, axs)
			ylim = axs.get_ylim()
			axs.set_ylim(ylim[0]*1.3-1.5,ylim[1]*1.3+1.5)
			# axs.set_ylabel('Change [%]')
			axs.set_title(m)
		plt.legend()
		plt.tight_layout()
		plt.savefig(f'{fname}_{name}.pdf')
		plt.clf()
	# plt.close('all')

def plotPareto(data, years, lineChange=True):
	def autolabel(rects, ax):
		"""Attach a text label above each bar in *rects*, displaying its height."""
		for rect in rects:
			width = rect.get_width()
			if width >= 0:
				ax.annotate(f'{width:.1f}', size ='small',
							xy=(-5, rect.get_y() + rect.get_height() / 2, ),
							ha='right', va='center')
			else:
				ax.annotate(f'{width:.1f}', size ='small',
							xy=(5, rect.get_y() + rect.get_height() / 2, ),
							ha='left', va='center')
							
	mode = ['incF', 'decF', 'incR', 'decR']
	fname = f'factorial_{years}y'
	if not lineChange: fname += '_noLineChange'
	folder = os.path.join('Figures_new2',fname)
	if not os.path.exists(folder):
		os.makedirs(folder)
	fname = os.path.join(folder,fname)	
	xlabel = ['B', 'L', 'D', 'CH', 'CL']
	for i, name in enumerate(['SAIFI','SAIDI','CAIFI','CAIDI','ASAI','AENS','ACCI']):
		fig, ax = plt.subplots(1,figsize=(4,8))
		percent = {}
		for j, m in enumerate(mode):
			A = data[i][0][j]
			if i == 4:
				A = [(1-A[i])/(1-A[0])*100-100 for i in range(1, len(A))]
			else:
				A = [A[i]/A[0]*100-100 for i in range(1, len(A))]
			for i,a in enumerate(A):
				percent[f'{xlabel[i]}:{m}:ICT'] = a

			B = data[i][1][j]
			if i == 4:
				B = [(1-B[i])/(1-B[0])*100-100 for i in range(1, len(B))]
			else:
				B = [B[i]/B[0]*100-100 for i in range(1, len(B))]
			for i,b in enumerate(B):
				percent[f'{xlabel[i]}:{m}:-no'] = b
		dataSorted = sorted(percent.items(), key=lambda x:x[1])
		plt.axvline(0, color='grey', linestyle='dashed', linewidth=1)
		color1 = next(ax._get_lines.prop_cycler)["color"]
		r1 = ax.barh(range(len(dataSorted)), [x[1] for x in dataSorted])
		ax.set(yticks=range(len(dataSorted)), yticklabels=[x[0][:-5] for x in dataSorted])
		autolabel(r1, ax)
		color2 = next(ax._get_lines.prop_cycler)["color"]
		for i,r in enumerate(r1):
			if dataSorted[i][0][-1] == 'o':
				r.set_color(color2)
		ax.set_xlabel('Change [%]')
		b_patch = mpatches.Patch(color=color1, label='ICT')
		g_patch = mpatches.Patch(color=color2, label='no-ICT')
		ax.legend(handles=[b_patch, g_patch], loc='lower right')
		plt.tight_layout()
		plt.savefig(f'{fname}_{name}_h.pdf')
		plt.clf()

def printFactorial(data, years):
	mode = ['incF', 'decF', 'incR', 'decR']
	xlabel = ['B', 'L', 'D', 'CH', 'CL']
	for i, name in enumerate(['SAIFI','SAIDI','CAIFI','CAIDI','ASAI','AENS','ACCI']):
		print(name)
		for j, m in enumerate(mode):
			A = data[i][0][j]
			print(m)
			print('ICT', A)
			if i == 4:
				A = [(1-A[i])/(1-A[0])*100-100 for i in range(1, len(A))]
			else:
				A = [A[i]/A[0]*100-100 for i in range(1, len(A))]

			B = data[i][1][j]
			print('no ICT', B)
			if i == 4:
				B = [(1-B[i])/(1-B[0])*100-100 for i in range(1, len(B))]
			else:
				B = [B[i]/B[0]*100-100 for i in range(1, len(B))]

def plotNoLineChange(years, s):
	xlabel = ['SAIFI','SAIDI','CAIFI','CAIDI','AENS','ACCI']
	data = []
	for j, m in enumerate(['', 'noLineChange_']):
		fname = f'{years}y_{m}'
		unavailability1 = np.loadtxt(os.path.join('Result',f'{s}ICT_{fname}unavailability.txt'), dtype=int).reshape((years, 66, 2))
		unavailability2 = np.loadtxt(os.path.join('Result',f'{s}noICT_{fname}unavailability.txt'), dtype=int).reshape((years, 66, 2))
		Udata1 = calcUnavailability(unavailability1)
		Udata2 = calcUnavailability(unavailability2)
		power1 = np.loadtxt(os.path.join('Result',f'{s}ICT_{fname}busENS.txt')).reshape((years, 2, 66, 2))
		power2 = np.loadtxt(os.path.join('Result',f'{s}noICT_{fname}busENS.txt')).reshape((years, 2, 66, 2))
		Pdata1 = calcPower(power1)
		Pdata2 = calcPower(power2)
		data.append(stats.runStats([Udata1[3],Udata2[3],Udata1[4],Udata2[4],Udata1[5],Udata2[5],Udata1[6],Udata2[6]], 
			['SAIFI','SAIDI','CAIFI','CAIDI']) + 
			stats.runStats([Pdata1[2],Pdata2[2],Pdata1[3],Pdata2[3]], ['AENS','ACCI']))
	
	percent = [[0 for _ in range(len(xlabel))] for _ in range(2)]
	change = [[0 for _ in range(len(xlabel))] for _ in range(2)]
	for i in range(len(data[0])):
		if i % 2:
			percent[1][int(i/2)] = data[1][i][0]/data[0][i][0]*100-100
			change[1][int(i/2)] = data[1][i][0]-data[0][i][0]
		else:
			percent[0][int(i/2)] = data[1][i][0]/data[0][i][0]*100-100
			change[0][int(i/2)] = data[1][i][0]-data[0][i][0]

	folder = os.path.join('Figures_new2',f'{s}{years}y')
	if not os.path.exists(folder):
		os.makedirs(folder)
	fname = os.path.join(folder,f'{s}{years}y')
	fig, ax = plt.subplots(figsize=(6,3))
	r1 = ax.bar(xlabel, percent[0], align='edge', width=-0.8/2, label='ICT')
	r2 = ax.bar(xlabel, percent[1], align='edge', width=0.8/2, label='no-ICT')
	plt.ylabel('Change from base case [%]')

	def autolabel(rects):
		"""Attach a text label above each bar in *rects*, displaying its height."""
		for rect in rects:
			height = rect.get_height()
			ax.annotate(f'{height:.0f}', xy=(rect.get_x() + rect.get_width() / 2, height),
						xytext=(0, 3),  # 3 points vertical offset
						textcoords="offset points", ha='center', va='bottom')
	autolabel(r1)
	autolabel(r2)
	ax.set_ylim(top=650)
	plt.legend()
	plt.tight_layout()
	plt.savefig(f'{fname}_noLineChange.pdf')

	fig, ax = plt.subplots(figsize=(6,3))
	r1 = ax.bar(xlabel, change[0], align='edge', width=-0.8/2, label='ICT')
	r2 = ax.bar(xlabel, change[1], align='edge', width=0.8/2, label='no-ICT')
	plt.ylabel('Change from base case')

	def autolabel2(rects):
		"""Attach a text label above each bar in *rects*, displaying its height."""
		for rect in rects:
			height = rect.get_height()
			ax.annotate(f'{height:.1f}', xy=(rect.get_x() + rect.get_width() / 2, height),
						xytext=(0, 3),  # 3 points vertical offset
						textcoords="offset points", ha='center', va='bottom')
	autolabel2(r1)
	autolabel2(r2)
	ax.set_ylim(top=6)
	plt.legend()
	plt.tight_layout()
	plt.savefig(f'{fname}_noLineChange_abs.pdf')
	plt.close('all')

def plotSeeds():
	fig, ax = plt.subplots(1, 3, figsize=(8,5))
	n = 2
	# choice = {0:'failureRate',1:'Un',2:'r',3:'SAIFI',4:'SAIDI',5:'CAIFI',6:'CAIDI',7:'ASAI'}
	choice = {0:'Total ENS P',1:'Bus ENS P',2:'AENS',3:'ACCI'}
	xlabel = ['67505', '10336', '33620']
	for i, years in enumerate([100, 1000, 10000]):
		data = []
		for s in xlabel:
			# Udata = calcUnavailability(np.loadtxt(os.path.join('Result',f's{s}_ICT_{years}y_unavailability.txt'), dtype=int).reshape((years, 66, 2)))
			Pdata = calcPower(np.loadtxt(os.path.join('Result',f's{s}_ICT_{years}y_busENS.txt')).reshape((years, 2, 66, 2)))
			data.append(Pdata[n])
		bplot = ax[i].boxplot(data, patch_artist=True, showfliers=True, showmeans=True, meanline=True)
		plt.setp(ax[i], xticks=[y + 1 for y in range(len(xlabel))], xticklabels=xlabel)
		plt.setp(bplot['means'], color='black')
		plt.setp(bplot['medians'], color='black')
		ax[i].set_xlabel('Seed')
		ax[i].set_title(f'{years} years')
	plt.tight_layout()
	plt.savefig(os.path.join('Figures_new2',f'box_seeds_{choice[n]}_ICT.pdf'))
	plt.close('all')

def CIprint(data, years):
	xlabel = ['SAIFI','SAIDI','CAIFI','CAIDI','ASAI','AENS','ACCI']
	cv = 3.29
	print(f'Confidence interval with critical value {cv:.2f}')
	for i, name in enumerate(xlabel):
		print(f'{name} & {data[i*2][0]-cv*data[i*2][1]/np.sqrt(years):.4f} & {data[i*2][0]+cv*data[i*2][1]/np.sqrt(years):.4f} &',
		      f'& {data[i*2+1][0]-cv*data[i*2+1][1]/np.sqrt(years):.4f} & {data[i*2+1][0]+cv*data[i*2+1][1]/np.sqrt(years):.4f} \\\\')

def plotAll(fname, years, power1, power2, unavailability1, unavailability2, linePlot=True, histPlot=True):
	folder = os.path.join('Figures_new2',fname)
	if not os.path.exists(folder):
		os.makedirs(folder)
	fname = os.path.join(folder,fname)
	choice = {0:'failureRate',1:'Un',2:'r',3:'SAIFI',4:'SAIDI',5:'CAIFI',6:'CAIDI',7:'ASAI',
	          8:'Total ENS P',9:'Bus ENS P',10:'AENS',11:'ACCI'}
	Pdata1 = calcPower(power1)
	Pdata2 = calcPower(power2)
	Udata1 = calcUnavailability(unavailability1)
	Udata2 = calcUnavailability(unavailability2)
	
	statData = stats.runStats([Udata1[3],Udata2[3],Udata1[4],Udata2[4],Udata1[5],Udata2[5],Udata1[6],Udata2[6],Udata1[7],Udata2[7]], list(choice.values())[3:8]) + stats.runStats([Pdata1[2],Pdata2[2],Pdata1[3],Pdata2[3]], list(choice.values())[10:])
	CIprint(statData, years)
	if histPlot:
		plotHist(Pdata1[2:], Pdata2[2:], ['AENS','ACCI'], fname)
		plotHist(Udata1[3:], Udata2[3:], ['SAIFI','SAIDI','CAIFI','CAIDI','ASAI'], fname)
	if linePlot:
		Pdata1 = runningAVGPower(Pdata1[0], Pdata1[1], Pdata1[2], Pdata1[3])
		Pdata2 = runningAVGPower(Pdata2[0], Pdata2[1], Pdata2[2], Pdata2[3])
		plotPower(Pdata1, Pdata2, fname)
		Udata1 = runningAVGUnavailability(Udata1[0], Udata1[1], Udata1[2], Udata1[3], Udata1[4], Udata1[5], Udata1[6], Udata1[7])
		Udata2 = runningAVGUnavailability(Udata2[0], Udata2[1], Udata2[2], Udata2[3], Udata2[4], Udata2[5], Udata2[6], Udata2[7])
		plotUnavailability(Udata1, Udata2, fname)

		plotVS([x[-1] for x in Udata1[0][0][:-1]], [x[-1] for x in Udata1[1][0][:-1]], r'$\lambda$', 'r', f'{fname}_ICT')

def loadGen():
	random.seed(452021)
	A = []
	B = []
	for i in range(65):
		A.append(random.random()*0.6+0.5)
		B.append(A[-1]*(random.random()*0.2+0.4))
	plt.figure(1, figsize=(7,3))
	plt.bar(range(1,1+len(A)), A, label='High')
	plt.bar(range(1,1+len(B)), B, label='Low')
	plt.xlabel('Node')
	plt.ylabel('Load [pu]')
	plt.legend(loc='lower right')
	plt.tight_layout()
	plt.savefig(os.path.join('Figures_new2',f'load.pdf'))
	for i in range(len(A)):
		print(A[i], ', ', B[i])

if __name__ == "__main__":
	loadGen()
	years = 10000
	lChange = ['', '_noLineChange'][1]
	fname = f'{years}y{lChange}'
	s = ['s67505_', 's10336_', 's33620_'][0]

	# data = calcFactorial(years, lineChange=False)
	# data = stats.import_numbers(years, lineChange=True)
	# printFactorial(data, years)
	# plotFactorial(data, years, lineChange=True)
	# plotPareto(data, years, lineChange=True)
	# plotSeeds()
	# plotNoLineChange(years, s)

	power1 = np.loadtxt(os.path.join('Result',f'{s}ICT_{fname}_busENS.txt')).reshape((years, 2, 66, 2))
	power2 = np.loadtxt(os.path.join('Result',f'{s}noICT_{fname}_busENS.txt')).reshape((years, 2, 66, 2))
	unavailability1 = np.loadtxt(os.path.join('Result',f'{s}ICT_{fname}_unavailability.txt'), dtype=int).reshape((years, 66, 2))
	unavailability2 = np.loadtxt(os.path.join('Result',f'{s}noICT_{fname}_unavailability.txt'), dtype=int).reshape((years, 66, 2))

	plotAll(f'{s}{fname}', years, power1, power2, unavailability1, unavailability2, linePlot=True, histPlot=True)
	print('That\'s all folks!')